/* basspitch~.c Version: 0.3
 * Author: worcs.edu
 * Date: March 2012
 */

#include <m_pd.h>
#include <math.h>

static t_class *basspitch_tilde_class;

typedef struct _basspitch_tilde {
  t_object  x_obj;
  t_sample f;
  t_outlet *freq;
/*t_outlet *fdebug;*/
  t_float lev_oc;
  t_sample filt_state[4];
  t_sample *buf;
  t_sample *pbuf;
  int	sc;
  int	trig;
} t_basspitch_tilde;

t_int *basspitch_tilde_perform(t_int *w)
{
    t_basspitch_tilde *x = (t_basspitch_tilde *)(w[1]);
    t_sample  *in_p	=    (t_sample *)(w[2]);
    int n			=           (int)(w[3]);

    t_sample *buf = x->buf;
    int i, len = n;
    t_sample *out, *in;
	float power = 0.;

	/* downsample by 1/16 (blocksize 64 ---> 4) */
	for (i=0; i<4; i++) 
    {
		t_sample state = x->filt_state[i];
		in = out = in_p;
        while( in < in_p+len ) 
        {
			t_sample samp = state + (*in++ * 0.5);
			state = *in++ * 0.25;
			*out++ = samp + state;
		}
		len=len/2;
		x->filt_state[i] = state;
	}
	
    in = in_p;
	while (len--) buf[x->sc++] = *in++;

	/* detect an attack in the correct freq range */
	if (!x->trig && x->sc >= 64) 
    {
		for (i=0; i<64; i++) 
       	{
			float t = (float) i/32 * M_PI;
			buf[i] *= cos(t) * -0.5 + 0.5;
       	}

       	mayer_realfft(64, buf);
		for (i=1; i<6; i++)
		{
			float weight = 1.;
			int sym = 64-i;
			if (i==1 || i==5)
				weight = 1.5;
			power +=  (buf[i] * buf[i] + buf[sym] * buf[sym]) * weight;
		}
		if (power > 50)
			x->trig = 1;

		x->sc = 0;
		goto done;
	}

	/* start main analysis */
	if (x->sc >= 512)
    {
		int bin = 0, binh = 0;
    	float fracbin = 0, maxl = 0, maxh = 0;

		/* save top 3/4 of buffer to temp buffer */
		out = x->pbuf;
		in = buf+128;
		len = 384;
		while (len--) *out++ = *in++;

		for (i=0; i<512; i++) 
        {
			float t = (float) i/256 * M_PI;
			buf[i] *= cos(t) * -0.5 + 0.5;
		}
		mayer_realfft(512, buf);

		if (x->trig == 1)
		{
			for (i=7; i<19; i++) 
    	    {
				int sym = 512-i;
				power = buf[i] * buf[i] + buf[sym] * buf[sym];
				buf[i] = power;
				if (power > maxl) 
            	{
					maxl = power;
					bin = i;
				}
			}
			for (i=19; i<44; i++) 
        	{
				int sym = 512-i;
				power = buf[i] * buf[i] + buf[sym] * buf[sym];
				buf[i] = power;
				if (power > maxh) 
            	{
					maxh = power;
					binh = i;
				}
			}

			if ( maxl < (maxh*(1 - x->lev_oc*0.40)) )
				bin = binh;
			fracbin = (buf[bin+1] - buf[bin-1])/(2*(2*buf[bin] - buf[bin-1] - buf[bin+1]));
			fracbin += bin;

			x->trig = bin;
			outlet_float(x->freq, sys_getsr()/(512*16) * fracbin);
		}
		/* check for sustain */
		else
		{
			int sym = 512 - x->trig;
			power = buf[x->trig] * buf[x->trig] + buf[sym] * buf[sym];
	/*DEBUG outlet_float(x->fdebug, power); */
			if (power < 5000)
			{
				x->trig = x->sc = 0;
				goto done;
			}
		}
		/* restore saved 3/4 to buffer*/
		out = buf;
		in = x->pbuf;
		len = 384;
		while (len--) *out++ = *in++;	
		x->sc = 384;
	}

done:	
  return (w+4);
}

void basspitch_tilde_dsp(t_basspitch_tilde *x, t_signal **sp)
{
	int n = sp[0]->s_n;
	if (n != 64)
	{
		error("basspitch: block size must be 64");
		return;
	}
	dsp_add(basspitch_tilde_perform, 3, x,
				sp[0]->s_vec, n);
}

void *basspitch_tilde_new(t_floatarg f)
{
  t_basspitch_tilde *x = (t_basspitch_tilde *)pd_new(basspitch_tilde_class);
  int i;

  if (f < 0 || f > 2)
  {
	error("basspitch: Octave compensation must be between 0 and 2");
	x->lev_oc = 0;
  }
  else
	x->lev_oc = f;
  x->buf = getbytes(512 * sizeof(t_sample));
  x->pbuf = getbytes(384 * sizeof(t_sample));
  x->trig = x->sc = 0;

  for(i=0; i<4; i++)
	x->filt_state[i] = 0;
 
  floatinlet_new (&x->x_obj, &x->lev_oc);
  x->freq = outlet_new(&x->x_obj, &s_float);
/*x->fdebug = outlet_new(&x->x_obj, &s_float);*/

  return (void *)x;
}

void basspitch_tilde_free(t_basspitch_tilde *x)
{
	freebytes(x->buf, 512 * sizeof(t_sample));
	freebytes(x->pbuf, 384 * sizeof(t_sample));
}

void basspitch_tilde_setup(void)
{
  basspitch_tilde_class = class_new(gensym("basspitch~"),
		(t_newmethod)basspitch_tilde_new,
		(t_method)basspitch_tilde_free, sizeof(t_basspitch_tilde),
		CLASS_DEFAULT, A_DEFFLOAT, 0);
  class_addmethod(basspitch_tilde_class,
		(t_method)basspitch_tilde_dsp, gensym("dsp"), 0);
  CLASS_MAINSIGNALIN(basspitch_tilde_class, t_basspitch_tilde, f);
  post ("basspitch~ version 0.3");
}

